@@ -1,11 +1,11 @@
 /*
-   This file is part of Callgrind, a Valgrind skin for call graph
+   This file is part of Callgrind, a Valgrind tool for call graph
    profiling programs.
 
-   Copyright (C) 2002-2007, Josef Weidendorfer (Josef.Weidendorfer@gmx.de)
+   Copyright (C) 2002-2012, Josef Weidendorfer (Josef.Weidendorfer@gmx.de)
 
-   This skin is derived from and contains lot of code from Cachegrind
-   Copyright (C) 2002 Nicholas Nethercote (njn25@cam.ac.uk)
+   This tool is derived from and contains lot of code from Cachegrind
+   Copyright (C) 2002-2012 Nicholas Nethercote (njn@valgrind.org)
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -41,7 +41,7 @@ static void print_indent(int s)
     /* max of 40 spaces */
     char sp[] = "                                        ";
     if (s>40) s=40;
-    VG_(printf)(sp+40-s);
+    VG_(printf)("%s", sp+40-s);
 }
 
 void CLG_(print_bb)(int s, BB* bb)
@@ -51,7 +51,7 @@ void CLG_(print_bb)(int s, BB* bb)
 	print_indent(s);
     }
 
-    VG_(printf)("BB %p (Obj '%s')", bb_addr(bb), bb->obj->name);
+    VG_(printf)("BB %#lx (Obj '%s')", bb_addr(bb), bb->obj->name);
 }
 
 static
@@ -112,7 +112,7 @@ void CLG_(print_execstate)(int s, exec_state* es)
 }
 
 
-void CLG_(print_bbcc)(int s, BBCC* bbcc, Bool jumpaddr)
+void CLG_(print_bbcc)(int s, BBCC* bbcc)
 {
   BB* bb;
 
@@ -129,55 +129,54 @@ void CLG_(print_bbcc)(int s, BBCC* bbcc, Bool jumpaddr)
   bb = bbcc->bb;
   CLG_ASSERT(bb!=0);
 
-#if 0 
-  if (jumpaddr)
-    VG_(printf)("%s +%p=%p, ",
+  VG_(printf)("%s +%#lx=%#lx, ",
 	      bb->obj->name + bb->obj->last_slash_pos,
-	      bb->jmp_offset, bb_jmpaddr(bb));
-  else
-#endif
-    VG_(printf)("%s +%p=%p, ",
-		bb->obj->name + bb->obj->last_slash_pos,
-		bb->offset, bb_addr(bb));
+	      bb->offset, bb_addr(bb));
   CLG_(print_cxt)(s+8, bbcc->cxt, bbcc->rec_index);
 }
 
 void CLG_(print_eventset)(int s, EventSet* es)
 {
-  int i;
+    int i, j;
+    UInt mask;
+    EventGroup* eg;
 
-  if (s<0) {
-    s = -s;
-    print_indent(s);
-  }
-
-  if (!es) {
-    VG_(printf)("(EventSet not set)\n");
-    return;
-  }
-
-  VG_(printf)("%5s (Size/Cap %d/%d): ",
-	      es->name, es->size, es->capacity);
-
-  if (es->size == 0)
-    VG_(printf)("-");
-  else {
-    for(i=0; i< es->size; i++) {
-      if (i>0) {
-	VG_(printf)(" ");
-	if (es->e[i-1].nextTop == i)
-	  VG_(printf)("| ");
-      }
-      VG_(printf)(es->e[i].type->name);
+    if (s<0) {
+	s = -s;
+	print_indent(s);
     }
-  }
-  VG_(printf)("\n");
+
+    if (!es) {
+	VG_(printf)("(EventSet not set)\n");
+	return;
+    }
+
+    VG_(printf)("EventSet %d (%d groups, size %d):",
+		es->mask, es->count, es->size);
+
+    if (es->count == 0) {
+	VG_(printf)("-\n");
+	return;
+    }
+
+    for(i=0, mask=1; i<MAX_EVENTGROUP_COUNT; i++, mask=mask<<1) {
+	if ((es->mask & mask)==0) continue;
+	eg = CLG_(get_event_group)(i);
+	if (!eg) continue;
+	VG_(printf)(" (%d: %s", i, eg->name[0]);
+	for(j=1; j<eg->size; j++)
+	    VG_(printf)(" %s", eg->name[j]);
+	VG_(printf)(")");
+    }
+    VG_(printf)("\n");
 }
 
 
 void CLG_(print_cost)(int s, EventSet* es, ULong* c)
 {
-  Int i, pos;
+    Int i, j, pos, off;
+    UInt mask;
+    EventGroup* eg;
 
     if (s<0) {
 	s = -s;
@@ -189,29 +188,36 @@ void CLG_(print_cost)(int s, EventSet* es, ULong* c)
       return;
     }
     if (!c) {
-      VG_(printf)("Cost (Null, EventSet %s)\n", es->name);
+      VG_(printf)("Cost (Null, EventSet %d)\n", es->mask);
       return;
     }
 
     if (es->size == 0) {
-      VG_(printf)("Cost (Nothing, EventSet %s with len 0)\n", es->name);
+      VG_(printf)("Cost (Nothing, EventSet with len 0)\n");
       return;
     } 
 
     pos = s;
-    pos += VG_(printf)("Cost %s [%p]: %s %llu", es->name, c, es->e[0].type->name, c[0]);
+    pos += VG_(printf)("Cost [%p]: ", c);
+    off = 0;
+    for(i=0, mask=1; i<MAX_EVENTGROUP_COUNT; i++, mask=mask<<1) {
+	if ((es->mask & mask)==0) continue;
+	eg = CLG_(get_event_group)(i);
+	if (!eg) continue;
+	for(j=0; j<eg->size; j++) {
 
-    i = 1;
-    while(i<es->size) {
-      if (pos > 70) {
-	VG_(printf)(",\n");
-	print_indent(s+5);
-	pos = s+5;
-      }
-      else
-	pos += VG_(printf)(", ");
-      pos += VG_(printf)("%s %llu", es->e[i].type->name, c[i]);
-      i++;
+	    if (off>0) {
+		if (pos > 70) {
+		    VG_(printf)(",\n");
+		    print_indent(s+5);
+		    pos = s+5;
+		}
+		else
+		    pos += VG_(printf)(", ");
+	    }
+
+	    pos += VG_(printf)("%s %llu", eg->name[j], c[off++]);
+	}
     }
     VG_(printf)("\n");
 }
@@ -220,13 +226,13 @@ void CLG_(print_cost)(int s, EventSet* es, ULong* c)
 void CLG_(print_short_jcc)(jCC* jcc)
 {
     if (jcc)
-	VG_(printf)("%p => %p [%llu/%llu,%llu,%llu]",
+	VG_(printf)("%#lx => %#lx [calls %llu/Ir %llu, Dr %llu, Dw %llu]",
 		    bb_jmpaddr(jcc->from->bb),
 		    bb_addr(jcc->to->bb),
 		    jcc->call_counter,
-		    jcc->cost ? jcc->cost[CLG_(sets).off_sim_Ir]:0,
-		    jcc->cost ? jcc->cost[CLG_(sets).off_sim_Dr]:0,
-		    jcc->cost ? jcc->cost[CLG_(sets).off_sim_Dw]:0);
+		    jcc->cost ? jcc->cost[fullOffset(EG_IR)]:0,
+		    jcc->cost ? jcc->cost[fullOffset(EG_DR)]:0,
+		    jcc->cost ? jcc->cost[fullOffset(EG_DW)]:0);
     else
 	VG_(printf)("[Skipped JCC]");
 }
@@ -243,10 +249,10 @@ void CLG_(print_jcc)(int s, jCC* jcc)
 	return;
     }
     VG_(printf)("JCC %p from ", jcc);
-    CLG_(print_bbcc)(s+9, jcc->from, True);
+    CLG_(print_bbcc)(s+9, jcc->from);
     print_indent(s+4);    
     VG_(printf)("to   ");
-    CLG_(print_bbcc)(s+9, jcc->to, False);
+    CLG_(print_bbcc)(s+9, jcc->to);
     print_indent(s+4);
     VG_(printf)("Calls %llu\n", jcc->call_counter);
     print_indent(s+4);
@@ -264,9 +270,9 @@ void CLG_(print_stackentry)(int s, int sp)
     }
 
     ce = CLG_(get_call_entry)(sp);
-    VG_(printf)("[%-2d] SP %p, RA %p", sp, ce->sp, ce->ret_addr);
+    VG_(printf)("[%-2d] SP %#lx, RA %#lx", sp, ce->sp, ce->ret_addr);
     if (ce->nonskipped)
-	VG_(printf)(" NonSkipped BB %p / %s",
+	VG_(printf)(" NonSkipped BB %#lx / %s",
 		    bb_addr(ce->nonskipped->bb),
 		    ce->nonskipped->cxt->fn[0]->name);
     VG_(printf)("\n");
@@ -295,7 +301,7 @@ void CLG_(print_bbcc_fn)(BBCC* bbcc)
 	return;
     }
 
-    VG_(printf)("%08x/%c  %d:", bb_addr(bbcc->bb), 
+    VG_(printf)("%08lx/%c  %d:", bb_addr(bbcc->bb), 
 		(bbcc->bb->sect_kind == Vg_SectText) ? 'T' :
 		(bbcc->bb->sect_kind == Vg_SectData) ? 'D' :
 		(bbcc->bb->sect_kind == Vg_SectBSS) ? 'B' :
@@ -334,14 +340,14 @@ void CLG_(print_bbcc_cost)(int s, BBCC* bbcc)
   bb = bbcc->bb;
   CLG_ASSERT(bb!=0);
     
-  CLG_(print_bbcc)(s, bbcc, False);
+  CLG_(print_bbcc)(s, bbcc);
 
   ecounter = bbcc->ecounter_sum;
 
   print_indent(s+2);
-  VG_(printf)("ECounter: sum %d ", ecounter);
+  VG_(printf)("ECounter: sum %llu ", ecounter);
   for(i=0; i<bb->cjmp_count; i++) {
-      VG_(printf)("[%d]=%d ",
+      VG_(printf)("[%d]=%llu ",
 		  bb->jmp[i].instr, bbcc->jmp[i].ecounter);
   }
   VG_(printf)("\n");
@@ -350,7 +356,7 @@ void CLG_(print_bbcc_cost)(int s, BBCC* bbcc)
   for(i=0; i<bb->instr_count; i++) {
       InstrInfo* ii = &(bb->instr[i]);
       print_indent(s+2);
-      VG_(printf)("[%2d] IOff %2d ecnt %3d ",
+      VG_(printf)("[%2d] IOff %2d ecnt %3llu ",
 		  i, ii->instr_offset, ecounter);
       CLG_(print_cost)(s+5, ii->eventset, bbcc->cost + ii->cost_offset);
 
@@ -370,23 +376,23 @@ void CLG_(print_addr)(Addr addr)
     Char fl_buf[FILENAME_LEN];
     Char fn_buf[FN_NAME_LEN];
     const UChar* obj_name;
-    SegInfo* si;
+    DebugInfo* di;
     int ln, i=0, opos=0;
 	
     if (addr == 0) {
-	VG_(printf)("%08x", addr);
+	VG_(printf)("%08lx", addr);
 	return;
     }
 
-    CLG_(get_debug_info)(addr, fl_buf, fn_buf, &ln, &si);
+    CLG_(get_debug_info)(addr, fl_buf, fn_buf, &ln, &di);
 
     if (VG_(strcmp)(fn_buf,"???")==0)
-	VG_(printf)("%p", addr);
+	VG_(printf)("%#lx", addr);
     else
-	VG_(printf)("%p %s", addr, fn_buf);
+	VG_(printf)("%#lx %s", addr, fn_buf);
 
-    if (si) {
-      obj_name = VG_(seginfo_filename)(si);
+    if (di) {
+      obj_name = VG_(DebugInfo_get_filename)(di);
       if (obj_name) {
 	while(obj_name[i]) {
 	  if (obj_name[i]=='/') opos = i+1;
@@ -429,10 +435,10 @@ void CLG_(print_context)(void)
   VG_(printf)("\n");
 }
 
-void* CLG_(malloc)(UWord s, char* f)
+void* CLG_(malloc)(HChar* cc, UWord s, char* f)
 {
-    CLG_DEBUG(3, "Malloc(%d) in %s.\n", s, f);
-    return VG_(malloc)(s);
+    CLG_DEBUG(3, "Malloc(%lu) in %s.\n", s, f);
+    return VG_(malloc)(cc,s);
 }
 
 #else /* CLG_ENABLE_DEBUG */
@@ -440,7 +446,7 @@ void* CLG_(malloc)(UWord s, char* f)
 void CLG_(print_bbno)(void) {}
 void CLG_(print_context)(void) {}
 void CLG_(print_jcc)(int s, jCC* jcc) {}
-void CLG_(print_bbcc)(int s, BBCC* bbcc, Bool b) {}
+void CLG_(print_bbcc)(int s, BBCC* bbcc) {}
 void CLG_(print_bbcc_fn)(BBCC* bbcc) {}
 void CLG_(print_cost)(int s, EventSet* es, ULong* cost) {}
 void CLG_(print_bb)(int s, BB* bb) {}
