@@ -7,7 +7,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2008-2009 OpenWorks LLP
+   Copyright (C) 2008-2012 OpenWorks LLP
       info@open-works.co.uk
 
    This program is free software; you can redistribute it and/or
@@ -54,9 +54,10 @@ Bool ML_(TyEnt__is_type)( TyEnt* te )
       case Te_EMPTY: case Te_INDIR: case Te_UNKNOWN: 
       case Te_Atom:  case Te_Field: case Te_Bound:
          return False;
-      case Te_TyBase:   case Te_TyPorR: case Te_TyTyDef:
-      case Te_TyStOrUn: case Te_TyEnum: case Te_TyArray:
-      case Te_TyFn:     case Te_TyQual: case Te_TyVoid:
+      case Te_TyBase:   case Te_TyPtr:     case Te_TyRef:
+      case Te_TyPtrMbr: case Te_TyRvalRef: case Te_TyTyDef:
+      case Te_TyStOrUn: case Te_TyEnum:    case Te_TyArray:
+      case Te_TyFn:     case Te_TyQual:    case Te_TyVoid:
          return True;
       default:
          vg_assert(0);
@@ -98,10 +99,15 @@ void ML_(pp_TyEnt)( TyEnt* te )
                      te->Te.Atom.value, te->Te.Atom.name);
          break;
       case Te_Field:
-         VG_(printf)("Te_Field(ty=0x%05lx,nLoc=%lu,loc=%p,\"%s\")",
-                     te->Te.Field.typeR, te->Te.Field.nLoc,
-                     te->Te.Field.loc,
-                     te->Te.Field.name ? te->Te.Field.name : (UChar*)"");
+         if (te->Te.Field.nLoc == -1)
+            VG_(printf)("Te_Field(ty=0x%05lx,pos.offset=%ld,\"%s\")",
+                        te->Te.Field.typeR, te->Te.Field.pos.offset,
+                        te->Te.Field.name ? te->Te.Field.name : (UChar*)"");
+         else
+            VG_(printf)("Te_Field(ty=0x%05lx,nLoc=%lu,pos.loc=%p,\"%s\")",
+                        te->Te.Field.typeR, te->Te.Field.nLoc,
+                        te->Te.Field.pos.loc,
+                        te->Te.Field.name ? te->Te.Field.name : (UChar*)"");
          break;
       case Te_Bound:
          VG_(printf)("Te_Bound[");
@@ -122,10 +128,20 @@ void ML_(pp_TyEnt)( TyEnt* te )
                      te->Te.TyBase.name ? te->Te.TyBase.name
                                         : (UChar*)"(null)" );
          break;
-      case Te_TyPorR:
-         VG_(printf)("Te_TyPorR(%d,%c,0x%05lx)",
-                     te->Te.TyPorR.szB,
-                     te->Te.TyPorR.isPtr ? 'P' : 'R',
+      case Te_TyPtr:
+         VG_(printf)("Te_TyPtr(%d,0x%05lx)", te->Te.TyPorR.szB,
+                     te->Te.TyPorR.typeR);
+         break;
+      case Te_TyRef:
+         VG_(printf)("Te_TyRef(%d,0x%05lx)", te->Te.TyPorR.szB,
+                     te->Te.TyPorR.typeR);
+         break;
+      case Te_TyPtrMbr:
+         VG_(printf)("Te_TyMbr(%d,0x%05lx)", te->Te.TyPorR.szB,
+                     te->Te.TyPorR.typeR);
+         break;
+      case Te_TyRvalRef:
+         VG_(printf)("Te_TyRvalRef(%d,0x%05lx)", te->Te.TyPorR.szB,
                      te->Te.TyPorR.typeR);
          break;
       case Te_TyTyDef:
@@ -232,9 +248,21 @@ void ML_(pp_TyEnt_C_ishly)( XArray* /* of TyEnt */ tyents,
          if (!ent->Te.TyBase.name) goto unhandled;
          VG_(printf)("%s", ent->Te.TyBase.name);
          break;
-      case Te_TyPorR:
+      case Te_TyPtr:
          ML_(pp_TyEnt_C_ishly)(tyents, ent->Te.TyPorR.typeR);
-         VG_(printf)("%s", ent->Te.TyPorR.isPtr ? "*" : "&");
+         VG_(printf)("*");
+         break;
+      case Te_TyRef:
+         ML_(pp_TyEnt_C_ishly)(tyents, ent->Te.TyPorR.typeR);
+         VG_(printf)("&");
+         break;
+      case Te_TyPtrMbr:
+         ML_(pp_TyEnt_C_ishly)(tyents, ent->Te.TyPorR.typeR);
+         VG_(printf)("*");
+         break;
+      case Te_TyRvalRef:
+         ML_(pp_TyEnt_C_ishly)(tyents, ent->Te.TyPorR.typeR);
+         VG_(printf)("&&");
          break;
       case Te_TyEnum:
          if (!ent->Te.TyEnum.name) goto unhandled;
@@ -392,29 +420,29 @@ Word ML_(TyEnt__cmp_by_cuOff_only) ( TyEnt* te1, TyEnt* te2 )
 
 /* Generates a total ordering on TyEnts based on everything except
    their .cuOff fields. */
-static __attribute__((always_inline)) Word UWord__cmp ( UWord a, UWord b ) {
+static inline Word UWord__cmp ( UWord a, UWord b ) {
    if (a < b) return -1;
    if (a > b) return 1;
    return 0;
 }
-static __attribute__((always_inline)) Word Long__cmp ( Long a, Long b ) {
+static inline Word Long__cmp ( Long a, Long b ) {
    if (a < b) return -1;
    if (a > b) return 1;
    return 0;
 }
-static __attribute__((always_inline)) Word Bool__cmp ( Bool a, Bool b ) {
+static inline Word Bool__cmp ( Bool a, Bool b ) {
    vg_assert( ((UWord)a) <= 1 );
    vg_assert( ((UWord)b) <= 1 );
    if (a < b) return -1;
    if (a > b) return 1;
    return 0;
 }
-static __attribute__((always_inline)) Word UChar__cmp ( UChar a, UChar b ) {
+static inline Word UChar__cmp ( UChar a, UChar b ) {
    if (a < b) return -1;
    if (a > b) return 1;
    return 0;
 }
-static __attribute__((always_inline)) Word Int__cmp ( Int a, Int b ) {
+static inline Word Int__cmp ( Int a, Int b ) {
    if (a < b) return -1;
    if (a > b) return 1;
    return 0;
@@ -476,8 +504,11 @@ Word ML_(TyEnt__cmp_by_all_except_cuOff) ( TyEnt* te1, TyEnt* te2 )
       if (r != 0) return r;
       r = UWord__cmp(te1->Te.Field.nLoc, te2->Te.Field.nLoc);
       if (r != 0) return r;
-      r = Bytevector__cmp(te1->Te.Field.loc, te2->Te.Field.loc,
-                          te1->Te.Field.nLoc);
+      if (te1->Te.Field.nLoc == -1)
+         r = Long__cmp(te1->Te.Field.pos.offset, te2->Te.Field.pos.offset);
+      else
+         r = Bytevector__cmp(te1->Te.Field.pos.loc, te2->Te.Field.pos.loc,
+                             te1->Te.Field.nLoc);
       return r;
    case Te_Bound:
       r = Bool__cmp(te1->Te.Bound.knownL, te2->Te.Bound.knownL);
@@ -495,12 +526,13 @@ Word ML_(TyEnt__cmp_by_all_except_cuOff) ( TyEnt* te1, TyEnt* te2 )
       if (r != 0) return r;
       r = Asciiz__cmp(te1->Te.TyBase.name, te2->Te.TyBase.name);
       return r;
-   case Te_TyPorR:
+   case Te_TyPtr:
+   case Te_TyRef:
+   case Te_TyPtrMbr:
+   case Te_TyRvalRef:
       r = Int__cmp(te1->Te.TyPorR.szB, te2->Te.TyPorR.szB);
       if (r != 0) return r;
       r = UWord__cmp(te1->Te.TyPorR.typeR, te2->Te.TyPorR.typeR);
-      if (r != 0) return r;
-      r = Bool__cmp(te1->Te.TyPorR.isPtr, te2->Te.TyPorR.isPtr);
       return r;
    case Te_TyTyDef:
       r = UWord__cmp(te1->Te.TyTyDef.typeR, te2->Te.TyTyDef.typeR);
@@ -568,14 +600,18 @@ void ML_(TyEnt__make_EMPTY) ( TyEnt* te )
          break;
       case Te_Field:
          if (te->Te.Field.name) ML_(dinfo_free)(te->Te.Field.name);
-         if (te->Te.Field.loc) ML_(dinfo_free)(te->Te.Field.loc);
+         if (te->Te.Field.nLoc > 0 && te->Te.Field.pos.loc)
+            ML_(dinfo_free)(te->Te.Field.pos.loc);
          break;
       case Te_Bound:
          break;
       case Te_TyBase:
          if (te->Te.TyBase.name) ML_(dinfo_free)(te->Te.TyBase.name);
          break;
-      case Te_TyPorR:
+      case Te_TyPtr:
+      case Te_TyRef:
+      case Te_TyPtrMbr:
+      case Te_TyRvalRef:
          break;
       case Te_TyTyDef:
          if (te->Te.TyTyDef.name) ML_(dinfo_free)(te->Te.TyTyDef.name);
@@ -652,7 +688,10 @@ MaybeULong ML_(sizeOfType)( XArray* /* of TyEnt */ tyents,
          if (ent2->tag == Te_UNKNOWN)
             return mk_MaybeULong_Nothing(); /*UNKNOWN*/
          return ML_(sizeOfType)( tyents, ent->Te.TyTyDef.typeR );
-      case Te_TyPorR:
+      case Te_TyPtr:
+      case Te_TyRef:
+      case Te_TyPtrMbr:
+      case Te_TyRvalRef:
          vg_assert(ent->Te.TyPorR.szB == 4 || ent->Te.TyPorR.szB == 8);
          return mk_MaybeULong_Just( ent->Te.TyPorR.szB );
       case Te_TyStOrUn:
@@ -683,6 +722,8 @@ MaybeULong ML_(sizeOfType)( XArray* /* of TyEnt */ tyents,
                                                   - bo->Te.Bound.boundL + 1) ));
          }
          return eszB;
+      case Te_TyVoid:
+         return mk_MaybeULong_Nothing(); /*UNKNOWN*/
       default:
          VG_(printf)("ML_(sizeOfType): unhandled: ");
          ML_(pp_TyEnt)(ent);
@@ -727,7 +768,10 @@ XArray* /*UChar*/ ML_(describe_type)( /*OUT*/PtrdiffT* residual_offset,
          case Te_TyEnum:
          case Te_TyFn:
          case Te_TyVoid:
-         case Te_TyPorR:
+         case Te_TyPtr:
+         case Te_TyRef:
+         case Te_TyPtrMbr:
+         case Te_TyRvalRef:
          case Te_TyBase:
             goto done;
 
@@ -747,26 +791,32 @@ XArray* /*UChar*/ ML_(describe_type)( /*OUT*/PtrdiffT* residual_offset,
                field = ML_(TyEnts__index_by_cuOff)(tyents, NULL, fieldR);
                vg_assert(field);
                vg_assert(field->tag == Te_Field);
-               vg_assert(field->Te.Field.loc);
-               vg_assert(field->Te.Field.nLoc > 0);
-               /* Re data_bias in this call, we should really send in
-                  a legitimate value.  But the expression is expected
-                  to be a constant expression, evaluation of which
-                  will not need to use DW_OP_addr and hence we can
-                  avoid the trouble of plumbing the data bias through
-                  to this point (if, indeed, it has any meaning; from
-                  which DebugInfo would we take the data bias? */
-               res = ML_(evaluate_Dwarf3_Expr)(
-                       field->Te.Field.loc, field->Te.Field.nLoc,
-                       NULL/*fbGX*/, NULL/*RegSummary*/,
-                       0/*data_bias*/,
-                       True/*push_initial_zero*/);
-               if (0) {
-                  VG_(printf)("QQQ ");
-                  ML_(pp_GXResult)(res);
-                  VG_(printf)("\n");
+               vg_assert(field->Te.Field.nLoc < 0
+                         || (field->Te.Field.nLoc > 0
+                             && field->Te.Field.pos.loc));
+               if (field->Te.Field.nLoc == -1) {
+                  res.kind = GXR_Addr;
+                  res.word = field->Te.Field.pos.offset;
+               } else {
+                  /* Re data_bias in this call, we should really send in
+                     a legitimate value.  But the expression is expected
+                     to be a constant expression, evaluation of which
+                     will not need to use DW_OP_addr and hence we can
+                     avoid the trouble of plumbing the data bias through
+                     to this point (if, indeed, it has any meaning; from
+                     which DebugInfo would we take the data bias? */
+                   res =  ML_(evaluate_Dwarf3_Expr)(
+                          field->Te.Field.pos.loc, field->Te.Field.nLoc,
+                          NULL/*fbGX*/, NULL/*RegSummary*/,
+                          0/*data_bias*/,
+                          True/*push_initial_zero*/);
+                  if (0) {
+                     VG_(printf)("QQQ ");
+                     ML_(pp_GXResult)(res);
+                     VG_(printf)("\n");
+                  }
                }
-               if (res.kind != GXR_Value)
+               if (res.kind != GXR_Addr)
                   continue;
                mul = ML_(sizeOfType)( tyents, field->Te.Field.typeR );
                if (mul.b != True)
