@@ -7,8 +7,8 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2005-2010 Nicholas Nethercote <njn@valgrind.org>
-   Copyright (C) 2005-2010 Cerion Armour-Brown <cerion@open-works.co.uk>
+   Copyright (C) 2005-2012 Nicholas Nethercote <njn@valgrind.org>
+   Copyright (C) 2005-2012 Cerion Armour-Brown <cerion@open-works.co.uk>
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -33,6 +33,7 @@
 #include "pub_core_basics.h"
 #include "pub_core_vki.h"
 #include "pub_core_vkiscnums.h"
+#include "pub_core_libcsetjmp.h"    // to keep _threadstate.h happy
 #include "pub_core_threadstate.h"
 #include "pub_core_aspacemgr.h"
 #include "pub_core_debuglog.h"
@@ -362,6 +363,7 @@ static SysRes do_clone ( ThreadId ptid,
       know that this thread has come into existence.  If the clone
       fails, we'll send out a ll_exit notification for it at the out:
       label below, to clean up. */
+   vg_assert(VG_(owns_BigLock_LL)(ptid));
    VG_TRACK ( pre_thread_ll_create, ptid, ctid );
 
    if (flags & VKI_CLONE_SETTLS) {
@@ -447,7 +449,6 @@ DECL_TEMPLATE(ppc64_linux, sys_ipc);
 DECL_TEMPLATE(ppc64_linux, sys_clone);
 //zz DECL_TEMPLATE(ppc64_linux, sys_sigreturn);
 DECL_TEMPLATE(ppc64_linux, sys_rt_sigreturn);
-//zz DECL_TEMPLATE(ppc64_linux, sys_sigaction);
 DECL_TEMPLATE(ppc64_linux, sys_fadvise64);
 
 PRE(sys_socketcall)
@@ -583,7 +584,7 @@ PRE(sys_socketcall)
       * (after all it's glibc providing the arguments array)
        PRE_MEM_READ( "socketcall.sendmsg(args)", ARG2, 3*sizeof(Addr) );
      */
-     ML_(generic_PRE_sys_sendmsg)( tid, ARG2_0, ARG2_1 );
+     ML_(generic_PRE_sys_sendmsg)( tid, "msg", (struct vki_msghdr *)ARG2_1 );
      break;
    }
 
@@ -594,7 +595,7 @@ PRE(sys_socketcall)
       * (after all it's glibc providing the arguments array)
        PRE_MEM_READ("socketcall.recvmsg(args)", ARG2, 3*sizeof(Addr) );
      */
-     ML_(generic_PRE_sys_recvmsg)( tid, ARG2_0, ARG2_1 );
+     ML_(generic_PRE_sys_recvmsg)( tid, "msg", (struct vki_msghdr *)ARG2_1 );
      break;
    }
 
@@ -700,7 +701,7 @@ POST(sys_socketcall)
     break;
 
   case VKI_SYS_RECVMSG:
-    ML_(generic_POST_sys_recvmsg)( tid, ARG2_0, ARG2_1 );
+    ML_(generic_POST_sys_recvmsg)( tid, "msg", (struct vki_msghdr *)ARG2_1, RES );
     break;
 
   default:
@@ -1046,7 +1047,7 @@ PRE(sys_rt_sigreturn)
    /* See comments on PRE(sys_rt_sigreturn) in syswrap-amd64-linux.c for
       an explanation of what follows. */
 
-   ThreadState* tst;
+   //ThreadState* tst;
    PRINT("sys_rt_sigreturn ( )");
 
    vg_assert(VG_(is_valid_tid)(tid));
@@ -1055,7 +1056,7 @@ PRE(sys_rt_sigreturn)
 
    ///* Adjust esp to point to start of frame; skip back up over handler
    //   ret addr */
-   tst = VG_(get_ThreadState)(tid);
+   //tst = VG_(get_ThreadState)(tid);
    //tst->arch.vex.guest_ESP -= sizeof(Addr);
    // Should we do something equivalent on ppc64-linux?  Who knows.
 
@@ -1076,73 +1077,6 @@ PRE(sys_rt_sigreturn)
    *flags |= SfPollAfter;
 }
 
-//zz /* Convert from non-RT to RT sigset_t's */
-//zz static 
-//zz void convert_sigset_to_rt(const vki_old_sigset_t *oldset, vki_sigset_t *set)
-//zz {
-//zz    VG_(sigemptyset)(set);
-//zz    set->sig[0] = *oldset;
-//zz }
-//zz PRE(sys_sigaction)
-//zz {
-//zz    struct vki_sigaction new, old;
-//zz    struct vki_sigaction *newp, *oldp;
-//zz 
-//zz    PRINT("sys_sigaction ( %d, %p, %p )", ARG1,ARG2,ARG3);
-//zz    PRE_REG_READ3(int, "sigaction",
-//zz                  int, signum, const struct old_sigaction *, act,
-//zz                  struct old_sigaction *, oldact);
-//zz 
-//zz    newp = oldp = NULL;
-//zz 
-//zz    if (ARG2 != 0) {
-//zz       struct vki_old_sigaction *sa = (struct vki_old_sigaction *)ARG2;
-//zz       PRE_MEM_READ( "sigaction(act->sa_handler)", (Addr)&sa->ksa_handler, sizeof(sa->ksa_handler));
-//zz       PRE_MEM_READ( "sigaction(act->sa_mask)", (Addr)&sa->sa_mask, sizeof(sa->sa_mask));
-//zz       PRE_MEM_READ( "sigaction(act->sa_flags)", (Addr)&sa->sa_flags, sizeof(sa->sa_flags));
-//zz       if (ML_(safe_to_deref)(sa,sizeof(sa)) 
-//zz           && (sa->sa_flags & VKI_SA_RESTORER))
-//zz          PRE_MEM_READ( "sigaction(act->sa_restorer)", (Addr)&sa->sa_restorer, sizeof(sa->sa_restorer));
-//zz    }
-//zz 
-//zz    if (ARG3 != 0) {
-//zz       PRE_MEM_WRITE( "sigaction(oldact)", ARG3, sizeof(struct vki_old_sigaction));
-//zz       oldp = &old;
-//zz    }
-//zz 
-//zz    //jrs 20050207: what?!  how can this make any sense?
-//zz    //if (VG_(is_kerror)(SYSRES))
-//zz    //   return;
-//zz 
-//zz    if (ARG2 != 0) {
-//zz       struct vki_old_sigaction *oldnew = (struct vki_old_sigaction *)ARG2;
-//zz 
-//zz       new.ksa_handler = oldnew->ksa_handler;
-//zz       new.sa_flags = oldnew->sa_flags;
-//zz       new.sa_restorer = oldnew->sa_restorer;
-//zz       convert_sigset_to_rt(&oldnew->sa_mask, &new.sa_mask);
-//zz       newp = &new;
-//zz    }
-//zz 
-//zz    SET_STATUS_from_SysRes( VG_(do_sys_sigaction)(ARG1, newp, oldp) );
-//zz 
-//zz    if (ARG3 != 0 && SUCCESS && RES == 0) {
-//zz       struct vki_old_sigaction *oldold = (struct vki_old_sigaction *)ARG3;
-//zz 
-//zz       oldold->ksa_handler = oldp->ksa_handler;
-//zz       oldold->sa_flags = oldp->sa_flags;
-//zz       oldold->sa_restorer = oldp->sa_restorer;
-//zz       oldold->sa_mask = oldp->sa_mask.sig[0];
-//zz    }
-//zz }
-//zz 
-//zz POST(sys_sigaction)
-//zz {
-//zz    vg_assert(SUCCESS);
-//zz    if (RES == 0 && ARG3 != 0)
-//zz       POST_MEM_WRITE( ARG3, sizeof(struct vki_old_sigaction));
-//zz }
-
 #undef PRE
 #undef POST
 
@@ -1359,13 +1293,13 @@ static SyscallTableEntry syscall_table[] = {
    LINXY(__NR_sched_rr_get_interval,  sys_sched_rr_get_interval), // 161
    GENXY(__NR_nanosleep,         sys_nanosleep),          // 162
    GENX_(__NR_mremap,            sys_mremap),             // 163
-// _____(__NR_setresuid,         sys_setresuid),          // 164
+   LINX_(__NR_setresuid,         sys_setresuid),          // 164
 
    LINXY(__NR_getresuid,         sys_getresuid),          // 165
 // _____(__NR_query_module,      sys_query_module),       // 166
    GENXY(__NR_poll,              sys_poll),               // 167
 // _____(__NR_nfsservctl,        sys_nfsservctl),         // 168
-// _____(__NR_setresgid,         sys_setresgid),          // 169
+   LINX_(__NR_setresgid,         sys_setresgid),          // 169
 
    LINXY(__NR_getresgid,         sys_getresgid),          // 170
 // _____(__NR_prctl,             sys_prctl),              // 171
@@ -1379,7 +1313,7 @@ static SyscallTableEntry syscall_table[] = {
 // _____(__NR_rt_sigsuspend,     sys_rt_sigsuspend),      // 178
    GENXY(__NR_pread64,           sys_pread64),            // 179
 
-// _____(__NR_pwrite64,          sys_pwrite64),           // 180
+   GENX_(__NR_pwrite64,          sys_pwrite64),           // 180
    GENX_(__NR_chown,             sys_chown),              // 181
    GENXY(__NR_getcwd,            sys_getcwd),             // 182
    LINXY(__NR_capget,            sys_capget),             // 183
@@ -1495,6 +1429,9 @@ static SyscallTableEntry syscall_table[] = {
    LINX_(__NR_inotify_add_watch,  sys_inotify_add_watch), // 276
    LINX_(__NR_inotify_rm_watch,   sys_inotify_rm_watch),  // 277
 
+   LINX_(__NR_pselect6,          sys_pselect6),           // 280
+   LINXY(__NR_ppoll,             sys_ppoll),              // 281
+
    LINXY(__NR_openat,            sys_openat),             // 286
    LINX_(__NR_mkdirat,           sys_mkdirat),            // 287
    LINX_(__NR_mknodat,           sys_mknodat),            // 288
@@ -1510,7 +1447,7 @@ static SyscallTableEntry syscall_table[] = {
    LINX_(__NR_faccessat,         sys_faccessat),          // 298
    LINX_(__NR_set_robust_list,   sys_set_robust_list),    // 299
    LINXY(__NR_get_robust_list,   sys_get_robust_list),    // 300
-//   LINX_(__NR_move_pages,        sys_ni_syscall),        // 301
+   LINXY(__NR_move_pages,        sys_move_pages),        // 301
    LINXY(__NR_getcpu,            sys_getcpu),            // 302
    LINXY(__NR_epoll_pwait,       sys_epoll_pwait),       // 303
    LINX_(__NR_utimensat,         sys_utimensat),         // 304
@@ -1528,10 +1465,13 @@ static SyscallTableEntry syscall_table[] = {
    LINXY(__NR_dup3,              sys_dup3),             // 316
    LINXY(__NR_pipe2,             sys_pipe2),            // 317
    LINXY(__NR_inotify_init1,     sys_inotify_init1),    // 318
-   LINXY(__NR_perf_counter_open, sys_perf_counter_open),// 319
+   LINXY(__NR_perf_event_open,   sys_perf_event_open),  // 319
    LINXY(__NR_preadv,            sys_preadv),           // 320
    LINX_(__NR_pwritev,           sys_pwritev),          // 321
-   LINXY(__NR_rt_tgsigqueueinfo, sys_rt_tgsigqueueinfo) // 322
+   LINXY(__NR_rt_tgsigqueueinfo, sys_rt_tgsigqueueinfo),// 322
+
+   LINXY(__NR_process_vm_readv,  sys_process_vm_readv), // 351
+   LINX_(__NR_process_vm_writev, sys_process_vm_writev) // 352
 };
 
 SyscallTableEntry* ML_(get_linux_syscall_entry) ( UInt sysno )
