@@ -7,7 +7,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2005-2010 Apple Inc.
+   Copyright (C) 2005-2012 Apple Inc.
       Greg Parker  gparker@apple.com
 
    This program is free software; you can redistribute it and/or
@@ -32,6 +32,7 @@
 
 #include "pub_core_basics.h"
 #include "pub_core_vki.h"
+#include "pub_core_libcsetjmp.h"   // to keep _threadstate.h happy
 #include "pub_core_threadstate.h"
 #include "pub_core_aspacemgr.h"
 #include "pub_core_xarray.h"
@@ -402,6 +403,16 @@ void wqthread_hijack(Addr self, Addr kport, Addr stackaddr, Addr workitem,
    VG_(sigprocmask)(VKI_SIG_SETMASK, &blockall, NULL);
 
    if (reuse) {
+
+      /* For whatever reason, tst->os_state.pthread appear to have a
+         constant offset of 72 on 10.7, but zero on 10.6 and 10.5.  No
+         idea why. */
+#     if DARWIN_VERS <= DARWIN_10_6
+      UWord magic_delta = 0;
+#     elif DARWIN_VERS >= DARWIN_10_7
+      UWord magic_delta = 0x48;
+#     endif
+
       // This thread already exists; we're merely re-entering 
       // after leaving via workq_ops(WQOPS_THREAD_RETURN). 
       // Don't allocate any V thread resources.
@@ -411,8 +422,14 @@ void wqthread_hijack(Addr self, Addr kport, Addr stackaddr, Addr workitem,
       vg_assert(mach_thread_self() == kport);
 
       tst = VG_(get_ThreadState)(tid);
+
+      if (0) VG_(printf)("wqthread_hijack reuse %s: tid %d, tst %p, "
+                         "tst->os_state.pthread %#lx, self %#lx\n",
+                         tst->os_state.pthread == self ? "SAME" : "DIFF",
+                         tid, tst, tst->os_state.pthread, self);
+
       vex = &tst->arch.vex;
-      vg_assert(tst->os_state.pthread == self);
+      vg_assert(tst->os_state.pthread - magic_delta == self);
    }
    else {
       // This is a new thread.
