@@ -1,4 +1,3 @@
-/* -*- mode: C; c-basic-offset: 3; indent-tabs-mode: nil; -*- */
 #ifndef __PUB_CORE_DRD_H
 #define __PUB_CORE_DRD_H
 
@@ -19,7 +18,8 @@ Bool DRD_(is_suppressed)(const Addr a1, const Addr a2);
 Bool DRD_(is_any_suppressed)(const Addr a1, const Addr a2);
 void DRD_(mark_hbvar)(const Addr a1);
 Bool DRD_(range_contains_suppression_or_hbvar)(const Addr a1, const Addr a2);
-void DRD_(start_tracing_address_range)(const Addr a1, const Addr a2);
+void DRD_(start_tracing_address_range)(const Addr a1, const Addr a2,
+                                       const Bool persistent);
 void DRD_(stop_tracing_address_range)(const Addr a1, const Addr a2);
 Bool DRD_(is_any_traced)(const Addr a1, const Addr a2);
 void DRD_(suppression_stop_using_mem)(const Addr a1, const Addr a2);
