@@ -8,7 +8,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2008-2011 Evan Geller
+   Copyright (C) 2008-2012 Evan Geller
       gaze@bea.ms
 
    This program is free software; you can redistribute it and/or
@@ -32,7 +32,7 @@
 #ifndef __VKI_SCNUMS_ARM_LINUX_H
 #define __VKI_SCNUMS_ARM_LINUX_H
 
-// From linux-2.6.26.2/include/asm-arm/unistd.h
+// From linux-2.6/arch/arm/include/asm/unistd.h
 
 #define __NR_restart_syscall		  0
 #define __NR_exit			  1
@@ -412,6 +412,8 @@
 #define __NR_syncfs			373
 #define __NR_sendmmsg			374
 #define __NR_setns			375
+#define __NR_process_vm_readv		376
+#define __NR_process_vm_writev		377
 
 
 
